import onnxruntime
import numpy as np
import torch

from hufr.models.utils import argmax_with_threshold
from typing import Callable, List, Union
from transformers import AutoTokenizer
from hufr.convert import convert_token_preds
from transformers.configuration_utils import PretrainedConfig


def softmax(x, axis=2):
    # Ensure numerical stability by subtracting the maximum value
    # along the second dimension before applying the softmax
    e_x = np.exp(x - np.max(x, axis=axis, keepdims=True))

    # Calculate softmax along the second dimension
    return e_x / np.sum(e_x, axis=axis, keepdims=True)


class TokenClassificationTransformerONNX:
    def __init__(
        self,
        onnx_model_path: Union[str, None],
        tokenizer: Union[Callable, str, None],
        execution_providers: List = ["CPUExecutionProvider"],
        threshold=0.9,
        config: Union[PretrainedConfig, None] = None,
    ):
        sess_option = onnxruntime.SessionOptions()
        sess_option.optimized_model_filepath = onnx_model_path
        sess_option.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_BASIC
        )
        if isinstance(tokenizer, str):
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        elif tokenizer is None:
            self.tokenizer = None
        self.execution_providers = execution_providers
        self.session = None
        if onnx_model_path is not None:
            self.session = onnxruntime.InferenceSession(
                onnx_model_path, sess_option, providers=["CPUExecutionProvider"]
            )
        self.threshold = threshold
        self.config = config

    @classmethod
    def from_pretrained(
        cls,
        onnx_model_path: str,
        tokenizer: Union[Callable, str, None] = None,
        execution_providers: List[str] = ["CPUExecutionProvider"],
        threshold=0.9,
        config: Union[PretrainedConfig, None] = None,
    ):
        """
        Instantiate the class with a pre-trained model path.

        Args:
            model_path (str): Path or identifier of the pre-trained model.

        Returns:
            TokenClassificationTransformerONNX: An instance of the class with the pre-trained model.
        """
        model = cls(
            onnx_model_path=onnx_model_path,
            tokenizer=tokenizer,
            execution_providers=execution_providers,
            threshold=threshold,
            config=config,
        )
        return model

    def predict(self, texts: List[str]) -> List[float]:
        """Predicts the label for each text in the list.

        Args:
        ----
            texts (List[str]): List of text to predict.
        Returns:
        -------
            List[int]: List of predicted label probabilities.
        """
        # TO DO: Add offset to handle long input sequences
        if self.session is None:
            raise AttributeError(
                "Please instantiate an onnx model using the from_pretrained method."
            )
        inputs = self.tokenizer(
            texts,
            return_tensors="pt",
            truncation=True,
            padding="longest",
            max_length=self.tokenizer.model_max_length,
        )
        inputs["input_ids"] = inputs["input_ids"].cpu().numpy()
        inputs["attention_mask"] = inputs["attention_mask"].cpu().numpy()
        inputs["token_type_ids"] = inputs["token_type_ids"].cpu().numpy()
        logits = self.session.run(
            None,
            {
                "input_ids": inputs["input_ids"],
                "attention_mask": inputs["attention_mask"],
                "token_type_ids": inputs["token_type_ids"],
            },
        )[0]
        probs = softmax(logits)
        predictions = argmax_with_threshold(
            torch.tensor(probs),
            threshold=self.threshold,
            default=self.config.label2id.get("O", 0),
            dim=2,
        )
        predictions = [
            [self.config.id2label[x] for x in pred] for pred in predictions.tolist()
        ]
        predictions = convert_token_preds(inputs, predictions, tokenizer=self.tokenizer)
        if isinstance(texts, str):
            return predictions[0]

        return predictions



import torch


def argmax_with_threshold(tensor, threshold, default, dim=0):
    """
    Helper function to get the maximum indices of a tensor along a given dimension while using a
    threshold value. If all tensor elements are below the threshold value then return the value
    specified by the default.

    Args:
        tensor (torch.Tensor): Input tensor.
        threshold (float): Threshold value for filtering maximum values.
        default: Value to be used if all elements are below the threshold.
        dim (int, optional): Dimension along which to compute the argmax. Defaults to 0.

    Returns:
        torch.Tensor: Tensor containing the indices of maximum values. If all elements are below the
        threshold, the tensor is filled with the specified default value.

    Note:
        This function calculates the argmax and maximum values along the specified dimension.
        If all maximum values are below the threshold, the function replaces the indices with the
        specified default value.
    """
    max_indices = torch.argmax(tensor, dim=dim)
    max_values = torch.max(tensor, dim=dim).values

    below_threshold = max_values <= threshold
    max_indices[below_threshold] = default

    return max_indices



import onnxruntime as ort

from hufr.onnx import model2onnx
from hufr.constants import DEFAULT_MODEL
from hufr.models import TokenClassificationTransformerONNX
from typing import Union, List


def instantiate_onnx_model(
    model_path_or_name: str = DEFAULT_MODEL,
    onnx_output_path: str = ".tmp/model.onnx",
    tokenizer_path_or_name: Union[str, None] = None,
    execution_providers: Union[List, None] = None,
):
    """
    Instantiates a TokenClassificationTransformerONNX model from a pre-trained model path or name.

    Args:
        model_path_or_name (str, optional): Path or name of the pre-trained model. Defaults to DEFAULT_MODEL.
        onnx_output_path (str, optional): Path to save the ONNX model. Defaults to ".tmp/model.onnx".
        tokenizer_path_or_name (Union[str, None], optional): Path or name of the tokenizer. If None, the tokenizer
            will be loaded from the model. Defaults to None.
        execution_providers (Union[List, None], optional): List of execution providers for ONNXRuntime. If None, the
            available providers will be used. Defaults to None.
    Returns:
        TokenClassificationTransformerONNX: An instance of the class with the pre-trained model.
    """
    if tokenizer_path_or_name is None:
        tokenizer_path_or_name = model_path_or_name
    onnx_model_path, config = model2onnx(
        model_path_or_name,
        onnx_output_path=onnx_output_path,
        tokenizer_path_or_name=tokenizer_path_or_name,
    )
    onnx_model = TokenClassificationTransformerONNX.from_pretrained(
        onnx_model_path,
        tokenizer=tokenizer_path_or_name,
        execution_providers=ort.get_available_providers()
        if execution_providers is None
        else execution_providers,
        config=config,
    )
    return onnx_model



import os
import logging

from pathlib import Path
from transformers.onnx import FeaturesManager, export
from transformers import AutoTokenizer, AutoModelForTokenClassification
from transformers.configuration_utils import PretrainedConfig


def model2onnx(
    model_path_or_name: str,
    onnx_output_path: str = None,
    tokenizer_path_or_name: str = None,
) -> tuple[str, PretrainedConfig]:
    """
    Export a model to ONNX format.
        Args:
            model_path_or_name (str): Path or identifier of the model.
            tokenizer_path_or_name (str, optional): Path or identifier of the tokenizer.
                If not provided, the model_path_or_name will be used. Defaults to None.
        Returns:
            path to model output or model as bytes depending on output parameter
    """
    model = AutoModelForTokenClassification.from_pretrained(model_path_or_name)
    if os.path.exists(onnx_output_path):
        logging.info(f"Found onnx weights at {onnx_output_path}")
        return onnx_output_path, model.config
    if tokenizer_path_or_name is None:
        tokenizer_path_or_name = model_path_or_name
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path_or_name)
    onnx_output_path = Path(
        f"{Path(model_path_or_name).path}.onnx"
        if onnx_output_path is None
        else onnx_output_path
    )
    if not onnx_output_path.parent.exists():
        logging.info(
            f"Created parent directory to save onnx weights at {onnx_output_path.parent}"
        )
        os.mkdir(onnx_output_path.parent)

    _, model_onnx_config = FeaturesManager.check_supported_model_or_raise(model)
    onnx_config = model_onnx_config(model.config)
    export(
        model=model,
        output=Path(onnx_output_path),
        preprocessor=tokenizer,
        tokenizer=None,
        config=onnx_config,
        opset=12,
    )
    logging.info(f"Exported onnx weights to {onnx_output_path}")
    return str(onnx_output_path), model.config
